use std::borrow::Borrow;

use rowan::{
    Checkpoint, GreenNode, GreenNodeBuilder, GreenNodeData, GreenTokenData, Language, NodeOrToken,
};

use crate::parser::{
    ast::lossless::{lex::SyntaxKind::*, Lang},
    Span,
};

use self::parser_to_events::{to_events, Event};

use super::lex::{self, SyntaxKind};

pub mod parser_to_events {
    use chumsky::prelude::*;

    use crate::parser::ast::lossless::lex::{
        self,
        SyntaxKind::{self, *},
    };

    #[derive(Debug, PartialEq, Eq, Clone, Copy)]
    pub enum Event {
        StartNode(SyntaxKind),
        StartErr(SyntaxError),
        EatToken,
        FinishNode,
        FinishErr,
    }

    #[derive(Debug, PartialEq, Eq, Clone, Copy)]
    pub enum SyntaxError {
        Expected(SyntaxKind),
        AttrExpectedValue,
        /// guessed if there's a newline and attr on next line without comma
        /// should then suggest comma after attr
        ExpectedCommaBetweenAttrs,
    }

    pub fn to_events(tokens: &[(SyntaxKind, &str)]) -> Vec<Event> {
        let mut only_toks: Vec<SyntaxKind> = tokens.iter().map(|(t, _)| *t).collect();
        let res = parser().parse(&only_toks);
        res.unwrap()
    }

    macro_rules! padded {
        ($parser:expr) => {{
            let ws = one_of([WHITESPACE, NEWLINE])
                .to(Event::EatToken)
                .repeated()
                .collect::<Vec<Event>>();
            ws.then($parser)
                .then(ws)
                .map(|((mut before, mut c), mut after)| {
                    before.append(&mut c);
                    before.append(&mut after);
                    before
                })
        }};
    }

    pub fn parser<'toks>() -> impl Parser<'toks, &'toks [SyntaxKind], Vec<Event>> {
        let ws = one_of([WHITESPACE, NEWLINE])
            .to(Event::EatToken)
            .repeated()
            .collect::<Vec<Event>>();
        // let ws_without_newlines = just(WHITESPACE)
        //     .to(Event::EatToken)
        //     .repeated()
        //     .collect::<Vec<Event>>();
        let parenthesized = |c| {
            just(L_PAREN)
                .to(vec![Event::EatToken])
                .then(c)
                .then(just(R_PAREN).to(vec![Event::EatToken]))
                .map(|((mut before, mut c), mut after)| {
                    before.append(&mut c);
                    before.append(&mut after);
                    before
                })
        };

        let expr = recursive(|expr| {
            let lit = one_of([INT_NUM, FLOAT_NUM, STRING]).to(vec![
                Event::StartNode(EXPR),
                Event::EatToken,
                Event::FinishNode,
            ]);

            let atom = lit.clone().or(parenthesized(expr));

            let ident = just(IDENT).to(vec![Event::EatToken]);
            let instr_name = ident
                .clone()
                .map(|mut v| {
                    v.insert(0, Event::StartNode(INSTR_NAME));
                    v
                })
                .foldl(
                    ws.then(ident).repeated(),
                    |mut ident, (mut ws, mut next)| {
                        ident.append(&mut ws);
                        ident.append(&mut next);
                        ident
                    },
                )
                .map(|mut v| {
                    v.push(Event::FinishNode);
                    v
                });
            let instr = padded!(instr_name)
                .then(
                    atom.clone()
                        .map(|mut v| {
                            v.insert(0, Event::StartNode(INSTR_PARAMS));
                            v
                        })
                        .foldl(
                            ws.then(atom.clone()).repeated(),
                            |mut cur, (mut ws, mut next)| {
                                cur.append(&mut ws);
                                cur.append(&mut next);
                                cur
                            },
                        )
                        .map(|mut v| {
                            v.push(Event::FinishNode);
                            v
                        }),
                )
                .map(|(mut name, mut params)| {
                    name.insert(0, Event::StartNode(INSTR));
                    name.append(&mut params);
                    name.push(Event::FinishNode);
                    name
                });
            padded!(instr.or(lit).or(atom))
        });
        expr
        // .map(|(lit, mut ev)| lit.append(&mut ev));
    }
}

#[derive(PartialEq, Eq)]
pub struct Parse {
    pub green_node: GreenNode,
}

impl std::fmt::Debug for Parse {
    fn fmt(&self, f: &mut std::fmt::Formatter<'_>) -> std::fmt::Result {
        debug_print_green_node(NodeOrToken::Node(self.green_node.borrow()), f, 0)
    }
}

fn debug_print_green_node(
    node: NodeOrToken<&GreenNodeData, &GreenTokenData>,
    f: &mut std::fmt::Formatter<'_>,
    lvl: i32,
) -> std::fmt::Result {
    for _ in 0..lvl {
        f.write_str("    ")?;
    }

    match node {
        NodeOrToken::Node(n) => {
            writeln!(f, "{:?} {{", Lang::kind_from_raw(node.kind()));
            for c in n.children() {
                debug_print_green_node(c, f, lvl + 1)?;
            }
            for _ in 0..lvl {
                f.write_str("    ")?;
            }
            f.write_str("}\n")
        }
        NodeOrToken::Token(t) => {
            writeln!(f, "{:?} {:?};", Lang::kind_from_raw(t.kind()), t.text())
        }
    }
}

#[derive(Debug)]
struct Parser<'src> {
    tokens: Vec<(SyntaxKind, &'src str)>,
    builder: GreenNodeBuilder<'src>,
    errors: Vec<SyntaxError>,
}

#[derive(Debug, PartialEq, Eq)]
enum SyntaxError {
    Expected(SyntaxKind),
    AttrExpectedValue,
    /// guessed if there's a newline and attr on next line without comma
    /// should then suggest comma after attr
    ExpectedCommaBetweenAttrs,
}

pub fn parse(src: &str) -> Parse {
    let tokens = lex::lex(src);
    Parser {
        tokens,
        builder: GreenNodeBuilder::new(),
        errors: Vec::new(),
    }
    .parse()
}

impl Parser<'_> {
    fn parse(mut self) -> Parse {
        let evs = to_events(&self.tokens);
        self.builder.start_node(ROOT.into());

        self.tokens.reverse();

        for ev in evs {
            match ev {
                Event::StartNode(kind) => self.builder.start_node(kind.into()),
                Event::StartErr(SyntaxError) => todo!(),
                Event::EatToken => self.bump(),
                Event::FinishNode => self.builder.finish_node(),
                Event::FinishErr => todo!(),
            }
        }

        self.builder.finish_node();
        Parse {
            green_node: self.builder.finish(),
        }
    }

    /// Advance one token, adding it to the current branch of the tree builder.
    fn bump(&mut self) {
        let (kind, text) = self.tokens.pop().unwrap();
        self.builder.token(kind.into(), text);
    }
    fn syntax_err(&mut self, err: SyntaxError) {
        let (_, text) = self.tokens.pop().unwrap();
        self.builder.token(PARSE_ERR.into(), text);
        self.errors.push(err);
    }
    fn syntax_err_by_checkpoint(&mut self, checkpoint: Checkpoint, err: SyntaxError) {
        self.builder.start_node_at(checkpoint, PARSE_ERR.into());
        self.builder.finish_node();
        self.errors.push(err);
    }
    fn expected(&mut self, expected: SyntaxKind) {
        self.syntax_err(SyntaxError::Expected(expected))
    }
    /// Peek at the first unprocessed token
    fn current(&self) -> Option<SyntaxKind> {
        self.tokens.last().map(|(kind, _)| *kind)
    }
    fn next(&self) -> Option<SyntaxKind> {
        self.tokens
            .get(self.tokens.len() - 2)
            .map(|(kind, _)| *kind)
    }
}
